/* SimpleLatexParser.java - jfig.utils.SimpleLatexParser
 *
 * trivial TeX / LaTeX command parser
 *
 * 15.04.03 - some more TeX math, support umlauts in normal text, ...
 * 17.12.02 - try to handle subscripts and superscripts
 * 15.12.02 - handle some "special short commands" like \$ or \%
 * 13.12.02 - more code
 * 12.12.02 - tokenizer works, some math
 * 11.12.02 - first code
 *
 * (C) 2002, fnh
 */

package jfig.utils;

import java.awt.*;
import java.io.*;
import java.util.*;
import java.lang.reflect.*;

import jfig.canvas.FigTrafo2D;
import jfig.gui.FontCache;
import jfig.gui.ColorCache;
import jfig.objects.*;


/**
 * a class to parse and convert simple TeX/LaTeX-style strings into 
 * formatted FIG compound objects.
 *
 * the following things work right now:
 * most elementary commands <code>\$, \%, \/, "a..."s etc.</code>,
 * font selection (index, ptSize, color) 
 * basic formatting (rm, em, tt, itt, btt) 
 * color selection 
 * basic math commands, if corresponding Unicode chars available
 * basic spacing commands
 * subscripts and superscripts with grouped {arguments}
 * 
 * but these do not:
 * subscript / superscript without grouped arguments
 * macros with arguments,
 * user-defined macros,
 */
public class SimpleLatexParser {
  boolean debug = false;                 // activates tons of diagnostic output
  boolean dump_tokens = false;      // should the parser dump the input tokens?
  boolean display_boxes = false;    // shows the bounding box around each token
  
  ColorCache colorCache   = ColorCache.getColorCache();   // jfig color manager
  FontCache fontCache     = FontCache.getFontCache();      // jfig font manager
  int mathFontIndex       = FigAttribs.FONT_TIMES_ROMAN_ITALIC;
  int mathGreekFontIndex  = FigAttribs.FONT_HELVETICA_OBLIQUE;
  int mathSymbolFontIndex = FigAttribs.FONT_HELVETICA;

  StringBuffer token;                                          // current token
  Vector       tokens;            // vector of tokens found in the input string
  Vector       types;                   // types of each token in vector tokens
  Stack  stack        = new Stack();                  // holds the parser state

  Box master;                        // toplevel Box constructed by this parser
  Box current;                                            // current active Box
  int fontIndex;                                   // index of current FIG font
  int ptSize;                                        // pt size of current font
  int colorIndex;                                // color index of current font
  int italic_correction = 0;          // pixels to be used as italic correction


  final static int INVALID = 0;
  final static int TEXT    = 1;
  final static int COMMAND = 2;
  final static int MATH    = 3;
  final static int BEGINGROUP  = 4;
  final static int ENDGROUP  = 5;
  final static int CONTROL = 6;
  final static int BEGINMATH   = 7;
  final static int ENDMATH     = 8;
  final static int SUPERSCRIPT = 9;
  final static int SUBSCRIPT   = 10;
  final static int MATHSYMBOL  = 11;

  final static String[] _mode_names = {
    "invalid", "text", "command", "math", "begingroup", "endgroup", "control",
    "beginmath", "endmath", "superscript", "subscript", "mathsymbol",
  };

  public final int[][] subscriptPtSizes = {
    // { normal ptsize, corresponding subscript ptsize }
    { 9, 6 }, { 10, 6 }, { 11, 7}, { 12, 9 },
    { 15, 10 }, { 17, 12 }, { 20, 12 }, 
    { 24, 15 }, { 30, 20 }, { 40, 27 }
  };




  /* ********************************************************************** */
  /* ********************************************************************** */
  /* inner class Box                                                        */
  /* ********************************************************************** */
  /* ********************************************************************** */


  /**
   * like TeX, we try to abut aligned Box'es, each of which consists
   * of other Boxes in turn, with a single character as the basic
   * building block.
   * Boxes are aligned along their baselines, where the integer values
   * (x,y) specify the 75 dpi screen-coordinates of the origin for
   * a box, while the values of w, a, d specify the width, ascent,
   * and descent of the box.
   * The formatting should probably changed to use FIG 2400 dpi world-
   * coordinates instead of screen-coordinates, but the JDK 1.3+ font
   * rendering does not really improve with higher resolution :-(
   * <p>
   * Unlike TeX, we use the predefined (Postscript) FIG fonts and
   * assume the Unicode encoding.
   * For jfig, a Box may be mapped to a FigCompound that holds FigText
   * objects.
   */
  class Box {
    protected int x, y, w, a, d;
    protected Box next;

    public void move( int dx, int dy ) {
      x += dx; y += dy;
    }

    public int getX() { return x; }
    public int getY() { return y; }
    public int getW() { return w; }
    public int getA() { return a; }
    public int getD() { return d; }

    public void setX( int x ) { this.x = x; }
    public void setY( int y ) { this.y = y; }
    public void setW( int w ) { this.w = w; }
    public void setA( int a ) { this.a = a; }
    public void setD( int d ) { this.d = d; }

    public Box getNextBox() { return next; }

    public void setNextBox( Box box ) { next = box; }

    public void append( Box child ) {
      this.measure();
      this.next = child;
      child.setOrigin( getAbutmentPoint() );
    }

    public void measure() {
      ; // does nothing, subclasses override as necessary
    }

    public Rectangle getBounds() {
      int xmin = Math.min( x, x+w );
      int width = Math.max( x, x+w ) - xmin;
      return new Rectangle( xmin, y-a, width, d+a );
    }

    public Point getAbutmentPoint() {
      return new Point( x+w, y );
    }

    public void setOrigin( Point p ) {
      this.x = p.x; this.y = p.y;
      measure();
    }

    public String toString() {
      return (this.getClass().getName() 
       + "[(" + x + "," + y + ") " + w 
       + " (" + a + "," + d + ")]");
    }

    public void dump( java.io.PrintStream PS ) {
      PS.println( this.toString() );
      if (getNextBox() != null) getNextBox().dump( PS );
    }

    /**
     * construct a FIG object from this Box and all its child boxes
     * and put those as members into the specified FigCompound group object.
     * Check the display_boxes flag to decide whether to generate
     * bounding boxes, too.
     */
    public void convertToFig( FigCompound parent ) {
      ; // add a red rectangle for this box
      if (display_boxes) {
   parent.fastAddMember( 
     getFigRectangle( parent, ColorCache.RED, 190 ));
      }
      if (next != null) next.convertToFig( parent );
    }

    /** 
     * construct a FigRectangle at 2400dpi world-coordinates from
     * our Box coordinates, and put it as a member into the specified
     * FigCompound group object.
     */
    public FigRectangle 
    getFigRectangle( FigCompound parent, int colorIndex, int layer ) {
      Rectangle r = getBounds();
      Point p1 = new Point( 32*r.x, 32*r.y );
      Point p2 = new Point( 32*(r.x+r.width), 32*(r.y+r.height) );
      FigRectangle rect = new FigRectangle();
      rect.setPoints( new Point[] { p1, p2 } );
      rect.setTrafo( parent.getTrafo() );
      rect.setAttributes( new FigAttribs() );
      rect.setLineColor( colorCache.get( colorIndex ));
      rect.setFillColor( colorCache.get( colorIndex ));
      rect.getAttributes().currentLayer = layer ;
      return rect;
    }
  } /* end inner class Box */




  /* ********************************************************************** */
  /* ********************************************************************** */
  /* inner class SubSuperScriptBox                                          */
  /* ********************************************************************** */
  /* ********************************************************************** */


  /**
   * a Box that holds a sub- and a superscript Box
   */
  class SubSuperScriptBox extends Box {
    Box superscriptBox;
    Box subscriptBox;
    int sa, sd; // ascent and descent of child boxes
    int ptSize;
    int fontIndex;


    public SubSuperScriptBox( int fontIndex, int ptSize ) {
      this.x = this.y = 0;
      this.fontIndex = fontIndex;
      this.ptSize = ptSize;
      measure();
    }


    public void measure() {
      FontMetrics fm = fontCache.getFontMetrics( fontIndex, ptSize );
      a = fm.getAscent();
      d = fm.getDescent();

      int subSize = getSubscriptPtSize();
      FontMetrics sm = fontCache.getFontMetrics( fontIndex, subSize );
      sa = sm.getAscent();
      sd = sm.getDescent();
      int total = 2*sa +2*sd + sd; // 2x ascent, 2x descent, 1x separation

      w = 0;
    }


    public Point getAbutmentPoint() {
      // calculate width of sub and superscript childs, if any
      Box tmp = null;

      tmp = subscriptBox;
      Point subscriptCursor = new Point(0,0);
      while( tmp != null) {
        subscriptCursor = tmp.getAbutmentPoint();
        tmp = tmp.getNextBox();
      }
      tmp = superscriptBox;
      Point superscriptCursor = new Point(0,0);
      while( tmp != null) {
        superscriptCursor = tmp.getAbutmentPoint();
        tmp = tmp.getNextBox();
      }
      return new Point( Math.max(subscriptCursor.x, superscriptCursor.x), y );
    }


    public int getSubscriptPtSize() {
      for( int i=0; i < subscriptPtSizes.length; i++) {
        if (subscriptPtSizes[i][0] >= ptSize) { // match found
           return subscriptPtSizes[i][1];
        }
      }
      return subscriptPtSizes[subscriptPtSizes.length-1][1]; // largest we have
    }


    public Point getSuperscriptBasePoint() {
      int ymid = (y-a) + (a+d)/2; // y-middle of 'master' box
      int ytop = ymid - (2*sa+2*sd+sd)/2; // y-top of superscript box
      // return new Point( x+italic_correction, ytop+sa );
      return new Point( x, ytop+sa );
    }


    public Point getSubscriptBasePoint() {
      int ymid = (y-a) + (a+d)/2; // y-middle of 'master' box
      int ybot = ymid + (2*sa+2*sd+sd)/2; // y-bottom of subscript box
      return new Point( x, ybot-sd );
    }


    public Box getSubscriptBox() { return subscriptBox; }
    public Box getSuperscriptBox() { return superscriptBox; }

    public void setSubscriptBox( Box b ) { subscriptBox = b; }
    public void setSuperscriptBox( Box b ) { superscriptBox = b; }


    public String toString() {
      String s_this = (this.getClass().getName() 
          + "[(" + x + "," + y + ") " + w 
          + " (" + a + "," + d + ")]");
      String s_super = (superscriptBox!=null) ? superscriptBox.toString() : "";
      String s_sub   = (subscriptBox!=null) ? subscriptBox.toString() : "";

      return s_this 
       + "\nsuperscript box:\n" + s_super 
       + "\nsubscript box:" + s_sub 
       + "\n";
    }


    public void convertToFig( FigCompound parent ) {
      ; // does nothing for this box itself, but recurses into the child boxes
      if (display_boxes) {
        parent.fastAddMember( 
          getFigRectangle( parent, ColorCache.GREEN, 195 ));
      }

      if (next != null) next.convertToFig( parent );
      if (superscriptBox != null) superscriptBox.convertToFig( parent );
      if (subscriptBox != null) subscriptBox.convertToFig( parent );
    }

  } /* end inner class SubSuperScriptBox */




  /* ********************************************************************** */
  /* ********************************************************************** */
  /* inner class CharBox                                                    */
  /* ********************************************************************** */
  /* ********************************************************************** */


  /**
   * a Box that holds a single character or a string of characters 
   * of the specified font.
   */
  class CharBox extends Box {
    protected int fontIndex;
    protected int ptSize;
    protected int colorIndex;
    protected String s;

    public CharBox( String s, int fontIndex, int ptSize, int colorIndex ) {
      this.x = this.y = 0;
      this.s = s;
      this.fontIndex = fontIndex;
      this.colorIndex = colorIndex;
      this.ptSize = ptSize;
      measure();
    }


    /** 
     * measures our string "s" and updates w,a,d
     */
    public void measure() {
      FontMetrics fm = fontCache.getFontMetrics( fontIndex, ptSize );
      a = fm.getAscent();
      d = fm.getDescent();
      w = fm.stringWidth( s );
    }  


    public void append( Box child ) {
      this.measure();
      this.next = child;
      Point tmp = getAbutmentPoint();
      // try to handle italic correction
      boolean thisIsItalic =  ((this.fontIndex & 0x01) != 0);
      boolean childIsItalic = ((child instanceof CharBox) 
            && ((((CharBox)child).fontIndex&0x01) != 0));
      if (thisIsItalic && !childIsItalic) tmp.x += italic_correction;
      child.setOrigin( tmp );
    }


    public String toString() {
      return (this.getClass().getName() 
       + "[(" + x + "," + y + ") " + w 
       + " (" + a + "," + d + ")]"
       + "'" + s + "'" );
    }


    public void convertToFig( FigCompound parent ) {
      FigTrafo2D trafo = parent.getTrafo();
      FigAttribs attribs = new FigAttribs();
      Point origin = new Point( 32*x, 32*y ); // zoom 1:1 2400dpi/75dpi
      FigText tmp = new FigText( origin, s, attribs, trafo );
      tmp.setFont( fontIndex );
      tmp.setFontSize( ptSize );
      tmp.setColor( colorCache.get( colorIndex ));
      parent.fastAddMember( tmp );
      if (display_boxes) {
        parent.fastAddMember( 
          getFigRectangle( parent, ColorCache.BLUE, 200 ));
      }
      if (next != null) next.convertToFig( parent );
    }

  } // CharBox




/* ******************************************************************* */
/* ******************************************************************* */
/*          tokenizer and parser functions                             */
/* ******************************************************************* */
/* ******************************************************************* */


  /**
   * tokenize one line of text input using the TeX grammar.
   * This method currently recognizes and supports the following constructs:
   * <ul>
   * <li>whitespace, ' ' or '\t' or '\n'
   * <li>punctuation: '.' ':' ';' ','
   * <li>plain text, ASCII a-z, A-Z, 0-9, 
   * <li>these special commmands: <code>\$ \% \& \"</code>
   * <li>umlauts from german.sty: <code>"a "A "o "O "u "U "s</code>
   * <li>commands consisting of backslash+name, limited by non a-zA-Z
   * <li>math mode delimiter: '$'
   * <li>group start and end, '{' and '}'
   * </ul>
   */
  void tokenize( String line ) {
    tokens = new Vector();
    types  = new Vector();
    token  = new StringBuffer( "" );
    char c = 0;
    char z = 0; // result of decoding a "short command" like \$ or "s
    int mode = TEXT, push = INVALID;
    for( int i=0, N=line.length(); i<N; i++ ) {
      c = line.charAt(i);
      z = getShortCommand( i, line );
      if (z != 0) i++;
// System.out.println( "... i,c,mode: " + i + " " + c + " " + _mode_names[mode]);

      if (mode == TEXT) {
        if      (z != 0)   { // interpret z as a normal text character 
                             token.append( z );
                           }
        else if (c == '$') { // switch from text to math mode
                             checkAddToken( token, TEXT );
                             checkAddToken( "$",   BEGINMATH );
                             token = new StringBuffer( "" );
                             mode = MATH;
                           }
        else if (c == '{') { // finish current token 
                             checkAddToken( token, TEXT );
                             checkAddToken( "{", BEGINGROUP );
                             token = new StringBuffer( "" );
                           }
        else if (c == '}') { // finish current token 
                             checkAddToken( token, TEXT );
                             checkAddToken( "}", ENDGROUP );
                             token = new StringBuffer( "" );
                           }
        else if (c == '\\'){ // finish current token
                             checkAddToken( token, TEXT );
                             token = new StringBuffer( "\\" );
                             push = TEXT;
                             mode = COMMAND;
                           }
        else if (c == '%')   { // start of comment
                             break;
                           }
        else if (isTextToken(c)) {
                             token.append( c );
                           }
        else {
                             errorInTokenizer( i, mode, line );
                           }
      }
      else if (mode == MATH) {
        if      (z != 0)   { // interpret z as a normal text character 
                             token.append( z );
                           }
        else if (isDigit(c)){ // digits should be upright in math
                             checkAddToken( token, MATH );
                             checkAddToken( ""+c, MATHSYMBOL );
                             token = new StringBuffer( "" );
                           }
        else if (isTextToken(c)) 
                           {
                             token.append( c );
                           }
        else if (isWhitespace(c) || isPunctuation(c)) 
                           {
                             token.append( c );
                           }
        else if (c == '$') { // end of math mode
                             checkAddToken( token, MATH );
                             checkAddToken( "$", ENDMATH );
                             token = new StringBuffer( "" );
                             mode = TEXT;
                           }
        else if (c == '^') { // start of superscript
                             checkAddToken( token, MATH );
                             checkAddToken( "^", SUPERSCRIPT );
                             token = new StringBuffer( "" );
                           }
        else if (c == '_') { // start of subscript
                             checkAddToken( token, MATH );
                             checkAddToken( "_", SUBSCRIPT );
                             token = new StringBuffer( "" );
                           }
        else if (c == '{') { // finish current token 
                             checkAddToken( token, MATH );
                             checkAddToken( "{", BEGINGROUP );
                             token = new StringBuffer( "" );
                           }
        else if (c == '}') { // finish current token 
                             checkAddToken( token, MATH );
                             checkAddToken( "}", ENDGROUP );
                             token = new StringBuffer( "" );
                           }
        else if (c == '\\'){ // a command
                             checkAddToken( token, MATH );
                             token = new StringBuffer( "\\" );
                             push = MATH;
                             mode = COMMAND;
                           }
        else if (isMathSymbol(c)) {
                             checkAddToken( token, MATH );
                             checkAddToken( ""+c, MATHSYMBOL );
                             token = new StringBuffer( "" );
                             mode = MATH;
                           }
        else if (c == '%') { // start of comment
                             break;
                           }
        else               {
                             errorInTokenizer( i, mode, line );
                           }
      }
      else if (mode == COMMAND) {
        if      (z != 0)   { // interpret z as a normal text character,
                             // but first finish the current command token 
                             checkAddToken( token, COMMAND );
                             token = new StringBuffer( ""+z );
                             mode = push;
                           }
        else if (isAscii(c)) 
                           {
                             token.append( c );
                           }
        else if (c == '\\'){ // one command following another
                             checkAddToken( token, COMMAND );
                             token = new StringBuffer( "\\" );
                             mode = COMMAND;
                           }
        else if (c == '$') { // finish command token, change to math mode
                             checkAddToken( token, COMMAND );
                             if (push == MATH) { // second $: back to text
                               checkAddToken( "$", ENDMATH );
                               mode = TEXT;
                             }
                             else { // first $: switch to math
                               checkAddToken( "$", BEGINMATH );
                               mode = MATH;
                             }
                             token = new StringBuffer( "" );
                           }
        else if (c == '%') { // start of comment
                             break;
                           }
        else if (c == '^') { // only valid from pushed math mode
                             if (push != MATH) {
                               errorInTokenizer( i, mode, line );
                               continue;
                             }
                             // finish command, switch to pushed mode
                             checkAddToken( token, COMMAND );
                             checkAddToken( "^", SUPERSCRIPT );
                             token = new StringBuffer( "" );
                             mode = push;
                           }
        else if (c == '_') { // only valid from pushed math mode
                             if (push != MATH) {
                               errorInTokenizer( i, mode, line );
                               continue;
                             }
                             // finish command, switch to pushed mode
                             checkAddToken( token, COMMAND );
                             checkAddToken( "_", SUBSCRIPT );
                             token = new StringBuffer( "" );
                             mode = push;
                           }
        else if (c == '{') { // finish current token 
                             checkAddToken( token, COMMAND );
                             checkAddToken( "{", BEGINGROUP );
                             token = new StringBuffer( "" );
                             mode = push;
                           }
        else if (c == '}') { // finish current token 
                             checkAddToken( token, COMMAND );
                             checkAddToken( "}", ENDGROUP );
                             token = new StringBuffer( "" );
                             mode = push;
                           }
        else if (c == ' ') { // ignore space behind command string
                             checkAddToken( token, COMMAND );
                             token = new StringBuffer( "" );
                             mode = push;
                           }
        else if (isDigit(c) || isPunctuation(c) || isWhitespace(c)) 
                           { // finish command, switch to pushed mode
                             checkAddToken( token, COMMAND );
                             token = new StringBuffer( ""+c );
                             mode = push;
                           } 
        else if ((push == MATH) && isMathSymbol(c)) {
                             checkAddToken( token, COMMAND );
                             checkAddToken( ""+c, MATHSYMBOL );
                             token = new StringBuffer( "" );
                             mode = push;
                           }
        else               {
                             errorInTokenizer( i, mode, line );
                           }
      }
      else {

      }
     
    } // end for loop
    // check for an unfinished token
    if ((token != null) && (token.length() > 0)) checkAddToken( token, mode );
  }  // tokenize


  private char getShortCommand( int index, String line ) {
    try {
      char    c = line.charAt( index );
      boolean b = (c == '\\') || (c == '\"');
      if (!b) return (char) 0;

      char    c2 = line.charAt( index+1 );
      if (c == '\\') {
        switch( c2 ) {
          case ' ':   return ' '; // space
          case '$':   return '$';
          case '%':   return '%';
          case '&':   return '&';
          case '{':   return '{';
          case '}':   return '}';
          case '|':   return '|';
          case '_':   return '_';
          case '\"':  return '\"';
          default:    return (char) 0;
        }
      }
      else if (c == '\"') {
        switch( c2) {
          case 'a':   return '';
          case 'o':   return '';
          case 'u':   return '';
          case 'A':   return '';
          case 'O':   return '';
          case 'U':   return '';
          case 's':   return '';
          default :  return (char) 0;
        }
      }
      else {
        return (char) 0;
      }
    }
    catch( Exception e ) {
      return (char) 0; // if something bad happened, this was NOT a short command
    }
  }


  private void checkAddToken( Object token, int mode ) {
    if (token == null) return;
    String tmp = token.toString();
    if (tmp.length()>0) {
      tokens.add( token );
      types.add ( new Integer(mode) );
    }
  }


  private void errorInTokenizer( int pos, int mode, String line ) {
    System.out.println( "Error in tokenize, mode= " + _mode_names[mode] );
    System.out.println( line );
    for( int i=0; i<pos; i++ ) System.out.print( " " );
    System.out.println( "^" );
  }


  public boolean isTextToken( char c ) {
    return     isAscii(c) 
      || isUmlaut(c) 
      || isDigit(c) 
      || isWhitespace(c) 
      || isPunctuation(c);
  }


  public boolean isAscii( char c ) {
    return 
       ((c >= 'a')&&(c <= 'z')) 
     ||((c >= 'A')&&(c <= 'Z'));
  }

  public boolean isUmlaut( char c ) {
    return (  
        (c == '') || (c == '')
     || (c == '') || (c == '')
     || (c == '') || (c == '')
     || (c == '')
     );
  }

  public boolean isDigit( char c ) {
    return ((c >= '0') && (c <= '9'));
  }


  public boolean isWhitespace( char c ) {
    return     (c == ' ') 
      || (c == '\t');
  }


  public boolean isPunctuation( char c ) {
    return     (c == '.') 
      || (c == ',')
      || (c == ';')
      || (c == ':')
      || (c == '!')
      || (c == '?')
      || (c == '\'') // FIXME: requires handling in math mode...
      || (c == '-')
      || (c == '(')  // we allow braces for text mode, too
      || (c == ')')
      || (c == '[')  // and also brackets
      || (c == ']')
   ;
  }


  public boolean isMathSymbol( char c ) {
    return     (c == '=')
      || (c == '>')
      || (c == '<')
      || (c == '+')
      || (c == '-')
      || (c == '|')
      || (c == '*')
      || (c == '(')
      || (c == ')')
      || (c == '/');
  }


  public void dumpTokens() {
    int indent = 0;
    for( int i=0,N=tokens.size(); i<N; i++ ) {
      for( int j=0; j < indent; j++) System.out.print( " " );

      String s    = tokens.elementAt(i).toString();
      int    mode = ((Integer) types.elementAt(i)).intValue();
      System.out.println( s + "   " + _mode_names[mode] );
      indent += s.length();
    }
  }


  public Object getState() {
    return new Object[] {
      new Integer( fontIndex ),
      new Integer( ptSize ),
      new Integer( colorIndex ),
    };
  }


  public void setState( Object tmp ) {
    try {
      int i=0;
      Object[] state = (Object[]) tmp;

      fontIndex = ((Integer) state[i++]).intValue();
      ptSize    = ((Integer) state[i++]).intValue();
      colorIndex= ((Integer) state[i++]).intValue();
    }
    catch( Throwable t ) {
msg( "-E- internal in setState: tmp= " + tmp );
      t.printStackTrace();
    }
  }


  public void append( Box box ) {
    current.append( box );
    current = box;
  }



  /**
   * build a compound set of nested Boxes for the given array of tokens,
   * starting with the specified FIG font in the given size.
   */
  public Box buildBox() {
    master = new CharBox( "", fontIndex, ptSize, colorIndex );
    current = master;

    for( int i=0,N=tokens.size(); i<N; i++ ) {
      String token = tokens.elementAt(i).toString();
      int    type  = ((Integer) types.elementAt(i)).intValue();

      switch(type) {
        case INVALID:
             msg( "-E-internal: mode=INVALID, token=" + token );
             break;
        case TEXT:
             // as even JDK 1.4.x returns mighty inaccurate values
             // for fontMetrics.stringWidth(), we try to minimize
             // the error by creating one box per character...
             // append(new CharBox(token,fontIndex,ptSize,colorIndex));
             for( int j=0; j < token.length(); j++ ) {
               String sc = "" + token.charAt(j);
               append( new CharBox( sc, fontIndex, ptSize, colorIndex ));
             }
             break;
        case COMMAND:
             // finish previous token
             String command = token.substring(1);
             handleCommand( command );
             break;
        case MATH:
             // see comment for TEXT: 
             // append(new CharBox(token,fontIndex,ptSize,colorIndex));
             for( int j=0; j < token.length(); j++ ) {
               String sc = "" + token.charAt(j);
               append( new CharBox( sc, fontIndex, ptSize, colorIndex ));
             }
             break;
        case BEGINGROUP:    // begingroup
             stack.push( getState() );
             ; // nothing else yet
             break;
        case ENDGROUP:   // endgroup
             // check if this ends a sub/super script box (triple push)
             setState( stack.pop() );
      
             // check, whether this begin/endgroup was due to a
             // sub/superscript
             if (!stack.isEmpty()) {
               Object ssbox = stack.peek(); 
               if (ssbox instanceof SubSuperScriptBox) {
                 ssbox = stack.pop();
                 current = (SubSuperScriptBox) ssbox;
                 setState( stack.pop() );
               }
             }
             break;
        case CONTROL:  // control character, e.g.
             msg( "-W- control character, should not happen: " + token );
             break;
        case BEGINMATH: // this switches the font to Times Italic for now
             stack.push( getState() );
             fontIndex = mathFontIndex;
             break;
        case ENDMATH:
             setState( stack.pop() );
             break;
        case SUPERSCRIPT: 
             {
if (debug) System.out.println( "\n\n\n" );
             SubSuperScriptBox tmp;
             if (current instanceof SubSuperScriptBox) {
                tmp = (SubSuperScriptBox) current;
      
if (debug) System.out.println( "SUPERSCRIPT: '" + token + "' " + tmp );
                // superscript entry still free?
                if (tmp.getSuperscriptBox() == null) { // yes!
            ; // we use tmp
                }
                else { // no: need to append a new box
            tmp = new SubSuperScriptBox( fontIndex, ptSize );
            append( tmp );
                }
             }
             else {
               tmp = new SubSuperScriptBox( fontIndex, ptSize );
               append( tmp );
             }
             stack.push( getState() );
             stack.push( tmp ); // mark that this was a begin subscript
      
             ptSize = tmp.getSubscriptPtSize(); // small
             Point base = tmp.getSuperscriptBasePoint();
             Box subBox = new CharBox( "", fontIndex, ptSize, colorIndex);
             subBox.setOrigin( base );
             tmp.setSuperscriptBox( subBox );
      
if (debug) System.out.println( "MASTER: base=" + base + "\nbox= " + tmp );
if (debug) System.out.println( "new superscriptbox=" + subBox + " " + ptSize );
      
             current = subBox;
             }
             break;
        case SUBSCRIPT: 
if (debug) System.out.println( "\n\n\n" );
             // create a new sub/superscript box and mark this box
             // as the new sub/superscript parent,
             SubSuperScriptBox tmp;
             if (current instanceof SubSuperScriptBox) {
                tmp = (SubSuperScriptBox) current;
if (debug) System.out.println( "SUBSCRIPT: '" + token + "' " + tmp );
                // subscript entry still free?
                if (tmp.getSubscriptBox() == null) { // yes!
            ; // we use tmp
                }
                else { // no: need to append a new box
            tmp = new SubSuperScriptBox( fontIndex, ptSize );
            append( tmp );
                }
             }
             else {
               tmp = new SubSuperScriptBox( fontIndex, ptSize );
               append( tmp );
             }
             stack.push( getState() );
             stack.push( tmp ); // mark that this was a begin subscript
      
             ptSize = tmp.getSubscriptPtSize(); // small
             Point base = tmp.getSubscriptBasePoint();
             Box subBox = new CharBox( "", fontIndex, ptSize, colorIndex);
             subBox.setOrigin( base );
             tmp.setSubscriptBox( subBox );
      
if (debug) System.out.println( "MASTER: base=" + base + "\nbox= " + tmp );
if (debug) System.out.println( "new subscriptbox=" + subBox + " " + ptSize );
      
             current = subBox;
             break;
        case MATHSYMBOL:
             // should handle math symbols...
             append( new CharBox( token, mathSymbolFontIndex, 
                ptSize, colorIndex ));
             break;
        default:
             System.out.println( "UNKNOWN MODE: " + type + " " + token );
      }
    }
    return master;
  }
      
      
  public void handleCommand( String command ) {
    String methodname = "_" + command;
    try {
      Class[]  formals = new Class[0];
      Method   method  = getClass().getMethod( methodname, formals );

      Object[] args = new Object[0];
      method.invoke( this, args );
    }
    catch( InvocationTargetException ite  ) {
      Throwable t = ite.getTargetException();
      msg( "-E- handleCommand: " + t );
      // t.printStackTrace();
    }
    catch( NoSuchMethodException e ) {
      msg( "-W- Command undefined: " + command );
      // msg( "-E- no such method: " + e + ":" + methodname );
      // e.printStackTrace();
      append( new CharBox( command, fontIndex, ptSize, colorIndex ) );
    }
    catch( Exception e ) {
      msg( "-E- exception: " + e + " for command: " + command );
      e.printStackTrace();
    }
  }



/* ******************************************************************* */
/* ******************************************************************* */
/*          built-in TeX and LaTeX macros                              */
/* ******************************************************************* */
/* ******************************************************************* */

  public void FIXME( String s ) {
    System.out.println( "-W- function not (yet) supported: " + s );
  }


  public void _bf() { // switch to boldface font
    // fonts 0..31 are sorted as normal, italic, bold, bold-italic
    fontIndex = fontIndex | 0x02;
  }

  public void _em() { // emphasize current text
    // int italic = fontIndex & 0x02;
    fontIndex = fontIndex ^ 0x01;
  }

  public void _rm() { // back to normal upright font
    // int italic = fontIndex & 0x02;
    fontIndex = fontIndex & 0xFC;
  }

  /**
   * request roman font in math mode. In TeX, this command takes one argument,
   * but we just change the font index globally here. This works in both
   * TeX/LaTeX and jfig2 if you use the following idiom:  
   * <code>$math italic text {\mathrm{math roman}} more math italic$.</code>
   */
  public void _mathrm() { // roman font in math mode, currently the same as rm
    fontIndex = fontIndex & 0xFC;
  }

  public void _tt()  { fontIndex = FigAttribs.FONT_COURIER; }
  public void _btt() { fontIndex = FigAttribs.FONT_COURIER_BOLD; }
  public void _itt() { fontIndex = FigAttribs.FONT_COURIER_OBLIQUE; }

  public void _black()      { colorIndex = ColorCache.BLACK;  }
  public void _red()        { colorIndex = ColorCache.RED;    }
  public void _green()      { colorIndex = ColorCache.GREEN;  }
  public void _blue()       { colorIndex = ColorCache.BLUE;   }
  public void _cyan()       { colorIndex = ColorCache.CYAN;   }
  public void _magenta()    { colorIndex = ColorCache.MAGENTA;}
  public void _yellow()     { colorIndex = ColorCache.YELLOW; }
  public void _white()      { colorIndex = ColorCache.WHITE;  }
  public void _darkblue()   { colorIndex = ColorCache.BLUE4;  }
  public void _lightblue()  { colorIndex = ColorCache.LTBLUE; }
  public void _darkgreen()  { colorIndex = ColorCache.GREEN4; }
  public void _darkred()    { colorIndex = ColorCache.RED4;   }
  public void _darkpink()   { colorIndex = ColorCache.PINK4;  }
  public void _gold()       { colorIndex = ColorCache.GOLD;   }

  public void _fivept()       { ptSize =  5; }
  public void _sixpt()        { ptSize =  6; }
  public void _sevenpt()      { ptSize =  7; }
  public void _eightpt()      { ptSize =  8; }
  public void _ninept()       { ptSize =  9; }
  public void _tenpt()        { ptSize = 10; }
  public void _elevenpt()     { ptSize = 11; }
  public void _twelvept()     { ptSize = 12; }
  public void _thirteenpt()   { ptSize = 13; }
  public void _fourteenpt()   { ptSize = 14; }
  public void _fifteenpt()    { ptSize = 15; }
  public void _sixteenpt()    { ptSize = 16; }
  public void _seventeenpt()  { ptSize = 17; }
  public void _eightteenpt()  { ptSize = 18; }
  public void _nineteenpt()   { ptSize = 19; }
  public void _twentypt()     { ptSize = 20; }
  public void _twentyonept()  { ptSize = 21; }
  public void _twentytwopt()  { ptSize = 22; }
  public void _twentythreept(){ ptSize = 23; }
  public void _twentyfourpt() { ptSize = 24; }
  public void _twentyfivept() { ptSize = 25; }
  public void _twentysixpt()  { ptSize = 26; }
  public void _twentysevenpt(){ ptSize = 27; }
  public void _twentyeightpt(){ ptSize = 28; }
  public void _twentyninept() { ptSize = 29; }
  public void _thirtypt()     { ptSize = 30; }
  public void _thirtyfivept() { ptSize = 35; }
  public void _fourtypt()     { ptSize = 40; }
  public void _fourtyfivept() { ptSize = 45; }
  public void _fiftypt()      { ptSize = 50; }
  public void _fiftyfivept()  { ptSize = 55; }
  public void _sixtypt()      { ptSize = 60; }
  public void _seventypt()    { ptSize = 70; }
  public void _eightypt()     { ptSize = 80; }
  public void _ninetypt()     { ptSize = 90; }
  public void _hundredpt()    { ptSize =100; }

  public void _xpt()          { ptSize = 10; }
  public void _xipt()         { ptSize = 11; }
  public void _xiipt()        { ptSize = 12; }
  public void _xiiipt()       { ptSize = 13; }
  public void _xivpt()        { ptSize = 14; }
  public void _xvpt()         { ptSize = 15; }
  public void _xvipt()        { ptSize = 16; }
  public void _xviipt()       { ptSize = 17; }
  public void _xviiipt()      { ptSize = 18; }
  public void _xixpt()        { ptSize = 19; }
  public void _xxpt()         { ptSize = 20; }
  public void _xxipt()        { ptSize = 21; }
  public void _xxiipt()       { ptSize = 22; }
  public void _xxiiipt()      { ptSize = 23; }
  public void _xxivpt()       { ptSize = 24; }
  public void _xxvpt()        { ptSize = 25; }
  public void _xxvipt()       { ptSize = 26; }
  public void _xxviipt()      { ptSize = 27; }
  public void _xxxpt()        { ptSize = 30; }
  public void _xxxvpt()       { ptSize = 35; }
  public void _xlpt()         { ptSize = 40; }

  public void _footnotesize() { FIXME( "footnotesize" ); }
  public void _tiny()         { FIXME( "tiny"  ); }
  public void _small()        { FIXME( "small" ); }
  public void _normalsize()   { FIXME( "normalsize" ); }
  public void _large()        { FIXME( "large" ); }
  public void _Large()        { FIXME( "Large" ); }
  public void _LARGE()        { FIXME( "LARGE" ); }
  public void _huge()         { FIXME( "huge" ); }
  

  // some basic ASCII chars
  // 
  public void _hashsign()       { unicode( "\u0023" ); } // # hash sign
  public void _atsign()         { unicode( "\u0040" ); } // @ at sign
  public void _lbracket()       { unicode( "\u005B" ); } // \[
  public void _backslash()      { unicode( "\\"     ); } // jikes :-(
  public void _rbracket()       { unicode( "\u005D" ); } // ]
  public void _lessthan()       { unicode( "<" ); } // <
  public void _equals()         { unicode( "=" ); } // =
  public void _greaterthan()    { unicode( ">" ); } // >
  public void _proportional()   { unicode( "\u2243" ); } // -~
  public void _approxeq()       { unicode( "\u2245" ); } // =~

  public void _lbrace()         { unicode( "\u007B" ); } // {
  public void _verticalbar()    { unicode( "\u007C" ); } // |
  public void _rbrace()         { unicode( "\u007D" ); } // }
  public void _tildechar()      { unicode( "\u007E" ); } // ~


  // LaTeX Companion table 8.2 - math mode accents
  //
  public void _not()    { FIXME( "\\noti#1" ); } // slash through next symbol
  public void _pmb()    { FIXME( "\\pmb#1"  ); } // poor man's bold
  public void _frac()   { FIXME( "\\frac#1#2");} // fraction #1 over #2
  public void _hat()    { FIXME( "\\hat#1"  ); } // ^ accent over #1
  public void _check()  { FIXME( "\\check#1"); } // v accent over #1
  public void _acute()  { FIXME( "\\acute#1"); } //  accent over #1
  public void _grave()  { FIXME( "\\grave#1"); } // ` accent over #1
  public void _brave()  { FIXME( "\\brave#1"); } // u accent over #1
  public void _bar()    { FIXME( "\\bar#1"  ); } // _ accent over #1
  public void _vec()    { FIXME( "\\vec#1"  ); } // arrow accent over #1
  public void _dot()    { FIXME( "\\dot#1"  ); } // derivate dot over #1
  public void _ddot()   { FIXME( "\\ddot#1" ); } // double dot over #1
  public void _tilde()  { FIXME( "\\tilde#1"); } // tilde accent over #1

  // LaTeX Companion table 8.3 - greek letters
  //
  public void _alpha()  { greek( "\u03b1" ); }
  public void _beta()   { greek( "\u03b2" ); }
  public void _gamma()  { greek( "\u03b3" ); }
  public void _delta()  { greek( "\u03b4" ); }
  public void _epsilon(){ greek( "\u03b5" ); }
  public void _varepsilon() { greek( "\u02b5" ); } // rounded epsilon
  public void _zeta()   { greek( "\u03b6" ); }
  public void _eta()    { greek( "\u03b7" ); }
  public void _theta()  { greek( "\u03b8" ); }
  public void _vartheta(){greek( "\u03d1" ); }
  public void _iota()   { greek( "\u03b9" ); }
  public void _kappa()  { greek( "\u03ba" ); }
  public void _lambda() { greek( "\u03bb" ); }
  public void _mu()     { greek( "\u03bc" ); }
  public void _nu()     { greek( "\u03bd" ); }
  public void _xi()     { greek( "\u03be" ); }
  public void _greeko() { greek( "\u03bf" ); } // greek "o" same as latin "o"
  public void _pi()     { greek( "\u03c0" ); }
  public void _varpi()  { greek( "\u03d6" ); }
  public void _rho()    { greek( "\u03c1" ); }
  public void _varrho() { greek( "\u03f1" ); } 
  public void _sigma()  { greek( "\u03c3" ); }
  public void _varsigma(){greek( "\u03da" ); }
  public void _tau()    { greek( "\u03c4" ); }
  public void _upsilon(){ greek( "\u03c5" ); }
  public void _phi()    { greek( "\u03d5" ); }
  public void _varphi() { greek( "\u03c6" ); }
  public void _chi()    { greek( "\u03c7" ); }
  public void _psi()    { greek( "\u03c8" ); }
  public void _omega()  { greek( "\u03c9" ); }

  public void _Alpha()  { greek( "\u0391" ); }
  public void _Beta()   { greek( "\u0392" ); }
  public void _Gamma()  { greek( "\u0393" ); }
  public void _Delta()  { greek( "\u0394" ); }
  public void _Epsilon(){ greek( "\u0395" ); }
  public void _Zeta()   { greek( "\u0396" ); }
  public void _Eta()    { greek( "\u0397" ); }
  public void _Theta()  { greek( "\u0398" ); }
  public void _Iota()   { greek( "\u0399" ); }
  public void _Kappa()  { greek( "\u039a" ); }
  public void _Lambda() { greek( "\u039b" ); }
  public void _Mu()     { greek( "\u039c" ); }
  public void _Nu()     { greek( "\u039d" ); }
  public void _Xi()     { greek( "\u039e" ); }
  public void _Greeko() { greek( "\u039f" ); } // greek "o" same as latin "o"
  public void _Pi()     { greek( "\u03A0" ); }
  public void _Rho()    { greek( "\u03A1" ); }
  public void _Sigma()  { greek( "\u03A3" ); }
  public void _Tau()    { greek( "\u03A4" ); }
  public void _Upsilon(){ greek( "\u03A5" ); }
  public void _Phi()    { greek( "\u03A6" ); }
  public void _Chi()    { greek( "\u03A7" ); }
  public void _Psi()    { greek( "\u03A8" ); }
  public void _Omega()  { greek( "\u03A9" ); }

  // LaTeX Companion table 8.4 - binary operation symbols
  //
  public void _pm()             { unicode( "\u00b1" ); } // +- plusminus
  public void _cap()            { unicode( "\u2229" ); } // Schnittmenge
  public void _diamond()        { unicode( "\u22c4" ); } // diamond
  public void _oplus()          { unicode( "\u2295" ); } // circled plus(xor)

  public void _mp()             { unicode( "\u2213" ); } // -/+
  public void _cup()            { unicode( "\u222A" ); } // Vereinigungsmenge
  public void _bigtriangleup()  { unicode( "\u2206" ); } // 
  public void _ominus()         { unicode( "\u2296" ); } // circled minus

  public void _times()          { unicode( "\u00d7" ); } // large X mult sign
  public void _uplus()          { unicode( "\u228e" ); } // plus sign in cup
  public void _bigtriangledown(){ unicode( "\u2207" ); } // 
  public void _otimes()         { unicode( "\u2297" ); } // circled mult. sign

  public void _div()            { unicode( "\u00f7" ); } // ./. division sign
  public void _sqcap()          { unicode( "\u2293" ); } // squared cap
  public void _triangleleft()   { unicode( "\u22b2" ); } // 
  public void _oslash()         { unicode( "\u2298" ); } // circled slash

  public void _ast()            { unicode( "\u2217" ); } // asterisk
  public void _sqcup()          { unicode( "\u2294" ); } // squared cup
  public void _triangleright()  { unicode( "\u22b3" ); } // 
  public void _odot()           { unicode( "\u2299" ); } // circled dot
 
  public void _star()           { unicode( "\u22C6" ); } // 5-peak star
  public void _vee()            { unicode( "\u2228" ); } // OR
  public void _lhd()            { unicode( "\u22b2" ); } // same as tri.left
  public void _bigcirc()        { unicode( "\u25cb" ); } // unicode geometrical
  
  public void _circ()           { unicode( "\u2218" ); } // small circle
  public void _wedge()          { unicode( "\u2227" ); } // AND 
  public void _rhd()            { unicode( "\u22b3" ); } // same as tri.left
  public void _dagger()         { unicode( "\u2020" ); } // Unicode punctuation

  public void _bullet()         { unicode( "\u2219" ); }
  public void _setminus()       { unicode( "\u2216" ); } // similar to backslash
  public void _unlhd()          { unicode( "\u22b4" ); } 
  public void _ddagger()        { unicode( "\u2021" ); } // double dagger

  public void _cdot()           { unicode( "\u22C5" ); } // centered small dot
  public void _wr()             { unicode( "\u2240" ); }
  public void _unrhd()          { unicode( "\u22b5" ); } 
  public void _amalg()          { unicode( "\u2210" ); }

  // Latex Companion table 8.5 - relation symbols
  //
  public void _leq()            { unicode( "\u2264" ); } // less equal
  public void _le()             { unicode( "\u2264" ); } // less equal
  public void _geq()            { unicode( "\u2265" ); } // greater equal
  public void _equiv()          { unicode( "\u2261" ); } // equivalent 
  public void _models()         { unicode( "\u22a8" ); } // |=
  public void _prec()           { unicode( "\u227a" ); } // precedes

  public void _succ()           { unicode( "\u227b" ); } // successor
  public void _sim()            { unicode( "\u223c" ); } // similar to
  public void _perp()           { unicode( "\u22a5" ); } // _|_
  public void _preceq()         { unicode( "\u227c" ); } // 
  public void _succeq()         { unicode( "\u227d" ); } // 

  public void _simeq()          { unicode( "\u22d6" ); } // FIXME: not 100%
  public void _mid()            { unicode( "\u2223" ); } // |
  public void _ll()             { unicode( "\u00ab" ); } // or 226a  <<
  public void _gg()             { unicode( "\u00bb" ); } // or 226b  >> 
  public void _asymp()          { unicode( "\u224d" ); } 

  public void _parallel()       { unicode( "\u2225" ); } // ||
  public void _subset()         { unicode( "\u2282" ); } // 
  public void _supset()         { unicode( "\u2283" ); } // 
  public void _approx()         { unicode( "\u2248" ); } // ~~
  public void _bowtie()         { unicode( "\u22c8" ); } // |><|

  public void _subseteq()       { unicode( "\u2286" ); } // 
  public void _supseteq()       { unicode( "\u2287" ); } // 
  public void _cong()           { unicode( "\u2245" ); } // =~
  public void _Join()           { unicode( "\u22c8" ); } // |><|
  public void _sqsubset()       { unicode( "\u228f" ); } //

  public void _sqsupset()       { unicode( "\u2290" ); } //
  public void _neq()            { unicode( "\u2260" ); } // not equal =/=
  public void _smile()          { unicode( "\u203f" ); } // Unicode punctuation
  public void _sqsubseteq()     { unicode( "\u2291" ); } //
  public void _sqsupseteq()     { unicode( "\u2292" ); } //

  public void _doteq()          { unicode( "\u2250" ); } // =.
  public void _frown()          { unicode( "\u2040" ); } // Unicode punctuation
  public void _in()             { unicode( "\u2208" ); } // element
  public void _ni()             { unicode( "\u220B" ); } // element righttoleft
  public void _propto()         { unicode( "\u221d" ); } // -~

  // public void _equals()      { unicode( "="      ); } // =, defined above
  public void _vdash()          { unicode( "\u22a2" ); } // |-
  public void _dashv()          { unicode( "\u22a3" ); } // -|
  //public void _lessthan()     { unicode( "<"      ); } // <, defined above
  //public void _greaterthan()  { unicode( ">"      ); } // >, defined above


  // LaTeX Companion table 8.6 - arrow symbols
  //
  public void _leftarrow()      { unicode( "\u2190" ); } // short,single arrows 
  public void _longleftarrow()  { unicode( "\u2190" ); } // FIXME
  public void _uparrow()        { unicode( "\u2191" ); }

  public void _Leftarrow()      { unicode( "\u21D0" ); } // double arrows
  public void _LongLeftarrow()  { unicode( "\u21D0" ); } // FIXME
  public void _Uparrow()        { unicode( "\u21D1" ); }

  public void _rightarrow()     { unicode( "\u2192" ); }
  public void _longrightarrow() { unicode( "\u2192" ); } // FIXME
  public void _downarrow()      { unicode( "\u2193" ); }

  public void _Rightarrow()     { unicode( "\u21D2" ); }
  public void _Longrightarrow() { unicode( "\u21D2" ); } // FIXME
  public void _Downarrow()      { unicode( "\u21D3" ); }

  public void _leftrightarrow() { unicode( "\u2194" ); }
  public void _longleftrightarrow() { unicode( "\u2194" ); } // FIXME
  public void _updownarrow()    { unicode( "\u2195" ); }

  public void _Leftrightarrow() { unicode( "\u21D4" ); }
  public void _Longleftrightarrow() { unicode( "\u21D4" ); } // FIXME
  public void _Updownarrow()    { unicode( "\u21D5" ); }

  public void _mapsto()         { unicode( "\u21A6" ); }
  public void _longmapsto()     { unicode( "\u21A6" ); } // FIXME
  public void _nearrow()        { unicode( "\u2197" ); } // north east 

  public void _hookleftarrow()  { unicode( "\u21A9" ); }
  public void _hookrightarrow() { unicode( "\u21AA" ); }
  public void _searrow()        { unicode( "\u2198" ); }

  public void _leftharpoonup()  { unicode( "\u21BC" ); }
  public void _rightharpoonup() { unicode( "\u21C0" ); }
  public void _swarrow()        { unicode( "\u2199" ); }

  public void _leftharpoondown(){ unicode( "\u21BD" ); }
  public void _rightharpoondown(){unicode( "\u21C1" ); }
  public void _nwarrow()        { unicode( "\u2196" ); }


  // LaTeX Companion table 8.7 - miscellaneous symbols
  //
  public void _ldots()          { unicode( "\u2026" ); } // 3 dots (punctuation)
  public void _cdots()          { unicode( "\u22ef" ); } // 3 centr'd horiz dots
  public void _vdots()          { unicode( "\u22ee" ); } // 3 vertical dots
  public void _ddots()          { unicode( "\u22f1" ); } // 3 diagonal dots
  public void _aleph()          { unicode( "\u2135" ); } // letterlike symbols

  public void _prime()          { unicode( "\u2032" ); } //  punctuation
  public void _forall()         { unicode( "\u2200" ); } // forall quantor
  public void _infty()          { unicode( "\u221E" ); } // infinity
  public void _hbar()           { unicode( "\u210f" ); } // letterlike symbols
  public void _emptyset()       { unicode( "\u2205" ); } // 

  public void _exists()         { unicode( "\u2204" ); } // exists quantor
  public void _nabla()          { unicode( "\u2207" ); }
  public void _surd()           { unicode( "\u221a" ); } // root symbol
  public void _Box()            { unicode( "\u25a1" ); } // geometrical shapes
  public void _triangle()       { unicode( "\u2206" ); }

  public void _Diamond()        { unicode( "\u25ca" ); } // geometrical shapes
  public void _imath()          { unicode( "i" ); } // FIXME
  public void _jmath()          { unicode( "j" ); } // FIXME
  public void _ell()            { unicode( "\u2113" ); } // l from letterlike
  public void _neg()            { unicode( "\u00ac" ); } // negation -,

  public void _top()            { unicode( "\u22a4" ); } // 
  public void _flat()           { unicode( "FIXME" ); } // 
  public void _natural()        { unicode( "FIXME" ); } // 
  public void _sharp()          { unicode( "FIXME" ); } // 
  public void _wp()             { unicode( "\u2118" ); } // 

  public void _bot()            { unicode( "\u22a5" ); } // 
  public void _clubsuit()       { unicode( "\u2663" ); } // Kreuz
  public void _diamondsuit()    { unicode( "\u2666" ); } // Karo
  public void _heartsuit()      { unicode( "\u2665" ); } // Herz
  public void _spadesuit()      { unicode( "\u2660" ); } // Pik

  public void _mho()            { unicode( "\u2127" ); } // 
  public void _Re()             { unicode( "\u211c" ); } // 
  public void _Im()             { unicode( "\u2111" ); } // 
  public void _angle()          { unicode( "\u2220" ); } // 
  public void _partial()        { unicode( "\u2202" ); } // partial diff.



  // LaTeX Companion table 8.8 - variable sized symbols
  // 
  public void _sum()            { unicode( "\u2211" ); } // sum sign (Sigma)
  public void _prod()           { unicode( "\u220F" ); } // product sign
  public void _coprod()         { unicode( "\u2210" ); } // coproduct sign
  public void _int()            { unicode( "\u222B" ); } // integral sign
  public void _oint()           { unicode( "\u222E" ); } // Kreisintegral

  // FIXME: the following are not big yet...
  public void _bigcap()         { unicode( "\u2229" ); } // Schnittmenge
  public void _bigcup()         { unicode( "\u222A" ); } // Vereinigungsmenge
  public void _bigsqcup()       { unicode( "\u2294" ); } // squared cup
  public void _bigvee()         { unicode( "\u2228" ); } // OR
  public void _bigwedge()       { unicode( "\u2227" ); } // AND 

  public void _bigodot()        { unicode( "\u2299" ); } // circled dot
  public void _bigotimes()      { unicode( "\u2297" ); } // circled mult. sign
  public void _bigoplus()       { unicode( "\u2295" ); } // circled plus(xor)
  public void _biguplus()       { unicode( "\u228e" ); } // plus sign in cup


  // LaTeX Companion table 8.9 - log-like symbols
  // 
  public void _arccos()         { mathfunction( "arccos" ); }
  public void _arcsin()         { mathfunction( "arcsin" ); }
  public void _arctan()         { mathfunction( "arctan" ); }
  public void _arg()            { mathfunction( "arg"    ); }

  public void _atan()           { mathfunction( "atan"   ); } // not in LaTeX

  public void _cos()            { mathfunction( "cos"    ); }
  public void _cosh()           { mathfunction( "cosh"   ); }
  public void _cot()            { mathfunction( "cot"    ); }
  public void _coth()           { mathfunction( "coth"   ); }

  public void _csc()            { mathfunction( "csc"    ); }
  public void _deg()            { mathfunction( "deg"    ); }
  public void _det()            { mathfunction( "det"    ); }
  public void _dim()            { mathfunction( "dim"    ); }

  public void _exp()            { mathfunction( "exp"    ); }
  public void _gcd()            { mathfunction( "gcd"    ); }
  public void _hom()            { mathfunction( "hom"    ); }
  public void _inf()            { mathfunction( "inf"    ); }

  public void _ker()            { mathfunction( "ker"    ); }
  public void _lg()             { mathfunction( "lg"     ); }
  public void _lim()            { mathfunction( "lim"    ); } 
  public void _liminf()         { mathfunction( "liminf" ); } 

  public void _limsup()         { mathfunction( "limsup" ); }
  public void _ln()             { mathfunction( "ln"     ); }
  public void _log()            { mathfunction( "log"    ); } 
  public void _max()            { mathfunction( "max"    ); }

  public void _min()            { mathfunction( "min"    ); }
  public void _Pr()             { mathfunction( "Pr"     ); }
  public void _sec()            { mathfunction( "sec"    ); }
  public void _sin()            { mathfunction( "sin"    ); }

  public void _sinh()           { mathfunction( "sinh"   ); }
  public void _sup()            { mathfunction( "sup"    ); }
  public void _tan()            { mathfunction( "tan"    ); }
  public void _tanh()           { mathfunction( "tanh"   ); }


  // LaTeX Companion 8.10 - delimiters
  //
  //public void _uparrow()      { unicode( "\u2191" ); }
  //public void _Uparrow()      { unicode( "\u21D1" ); }
  //public void _downarrow()    { unicode( "\u2193" ); }
  //public void _Downarrow()    { unicode( "\u21D3" ); }

  //public void _lbrace()       { unicode( "\u007B" ); } // {
  //public void _rbrace()       { unicode( "\u007D" ); } // }
  //public void _updownarrow()  { unicode( "\u2195" ); }
  //public void _Updownarrow()  { unicode( "\u21D5" ); }

  public void _lfloor()         { unicode( "FIXME"  ); }
  public void _rfloor()         { unicode( "FIXME"  ); }
  public void _lceil()          { unicode( "\u02e5" ); }
  public void _rceil()          { unicode( "\u02e9" ); }

  public void _langle()         { unicode( "FIXME"  ); }
  public void _rangle()         { unicode( "FIXME"  ); }
  public void _slash()          { unicode( "/"      ); }
  // public void _backslash()   { unicode( "\\"     ); } // jikes :-(

  // public void _verticalbar() { unicode( "\u007C" ); } // |
  public void _dverticalbar()   { unicode( "\u2225" ); } // ||


  // LaTeX Companion table 8.11 - large delimiters
  //
  public void _rmoustache()     { unicode( "FIXME"  ); }
  public void _lmoustache()     { unicode( "FIXME"  ); }
  public void _rgroup()         { unicode( "FIXME"  ); }
  public void _lgroup()         { unicode( "FIXME"  ); }

  public void _arrowvert()      { unicode( "FIXME"  ); }
  public void _Arrowvert()      { unicode( "FIXME"  ); }
  public void _bracevert()      { unicode( "FIXME"  ); }


  // LaTeX Companion table 8.12 - math constructs
  //
  public void _widetilde()      { FIXME( "widetilde#1" ); }
  public void _widehat()        { FIXME( "widehat#1"   ); }
  public void _overleftarrow()  { FIXME( "overleftarrow#1"  ); }
  public void _overrightarrow() { FIXME( "overrightarrow#1" ); }

  public void _overline()       { FIXME( "overline#1"  ); }
  public void _underline()      { FIXME( "underline#1" ); }
  public void _overbrace()      { FIXME( "overbrace#1" ); }
  public void _underbrace()     { FIXME( "underbrace#1"); }


  public void _sqrt()           { FIXME( "sqrt#1"); }
  public void _sqrtn()          { FIXME( "sqrtn#1"); }
  // public void _frac()        { FIXME( "\\frac#1#2");} // see above 8.2



  // LaTeX Companion table 8.13 - AMS greek and hebrew
  //
  public void _digamma()        { unicode( "F" ); }
  public void _beth()           { unicode( "\u2136" ); }
  public void _daleth()         { unicode( "\u2138" ); }
  public void _gimel()          { unicode( "\u2137" ); }



  // LaTeX Companion table 8.14++ - AMS stuff (incomplete)
  // 
  public void _looparrowleft()  { unicode( "\u21AB" ); }
  public void _looparrowright() { unicode( "\u21AC" ); }

  public void _leftrightsquigarrow() {unicode( "\u21AD" ); }
  public void _circlearrowleft() {unicode( "\u21BA" ); }
  public void _circlearrowright(){unicode( "\u21BB" ); }

  public void _nleftarrow()     { unicode( "\u219A" ); }
  public void _nrightarrow()    { unicode( "\u219B" ); }
  public void _nleftrightarrow(){ unicode( "\u21AE" ); }
  public void _nLeftarrow()     { unicode( "\u21CD" ); }
  public void _nLeftrightarrow(){ unicode( "\u21CE" ); }
  public void _nRightarrow()    { unicode( "\u21CF" ); }



  public void _define()         { unicode( "\u225D" ); } // =def=
  public void _lll()            { unicode( "\u22D8" ); } // <<<
  public void _ggg()            { unicode( "\u22D9" ); } // >>> 
  public void _cdotvar()        { unicode( "\u00B7" ); } // centered small dot
  public void _minus()          { unicode( "\u2212" ); } // large minus sign
  



  // currency stuff
  public void _euro()           { unicode( "\u20AC" ); } // Euro
  public void _pound()          { unicode( "\u20A4" ); } // britisch pound
  public void _pts()            { unicode( "\u20A7" ); } // spanisch pesetas
  public void _percent()        { unicode( "%"      ); } // percent sign
  public void _promille()       { unicode( "\u2030" ); } // 0/00 promille sign

  public void _copyright()      { unicode( "\u00a9" ); } // (C) sign
  public void _paragraph()      { unicode( "\u00b6" ); } // paragraph P sign
  public void _degrees()        { unicode( "\u00b0" ); } //  sign
  

  // Unicode "letterlike symbols"
  public void _setZ()           { unicode( "\u2124" ); } // ZZ
  public void _setN()           { unicode( "\u2115" ); } // IN
  public void _setR()           { unicode( "\u211D" ); } // IR
  public void _celsius()        { unicode( "\u2103" ); } // degrees celsius, oC
  public void _fahrenheit()     { unicode( "\u2109"  );} // degrees F, oF

  // Unicode "math", 2200 .. 22FF
  public void _complement()     { unicode( "\u2201" ); } // big C
  public void _nexists()        { unicode( "\u2204" ); } // exists not
  public void _varnothing()     { unicode( "\u2205" ); } // empty set
  // public void _triangle()    { unicode( "\u2206" ); } // defined in 8.4 above


  public void _notElementIn()   { unicode( "\u2209" ); }
  public void _mathEpsilon()    { unicode( "\u220A" ); }
  public void _square()         { unicode( "\u22A1" ); } // square

  //                                                     // ./+
  public void _mathslash()      { unicode( "\u2215" ); }
  public void _root()           { unicode( "\u221A" ); }
  public void _cubicroot()      { unicode( "\u221B" ); }
  public void _fourthroot()     { unicode( "\u221C" ); }
  public void _varpropto()      { unicode( "\u221D" ); } // "half" infinity

  public void _iint()           { unicode( "\u222C" ); } // double integral
  public void _iiint()          { unicode( "\u222D" ); } // triple integral




  public void _telephone()      { unicode( "\u260E" ); } // Dingbats phone
  public void _smiley()         { unicode( "\u263A" ); } // 
  public void _smileyblack()    { unicode( "\u263B" ); } // 
  public void _male()           { unicode( "\u2642" ); } // o/
  public void _female()         { unicode( "\u2640" ); } // +o


  // spacing
  public void _thinspace() {
    Box tmp = new Box(); tmp.setW( 5 ); append( tmp );
  }

  public void _qquad() {

  }
  public void _emspace() {
    CharBox M = new CharBox( "M", fontIndex, ptSize, colorIndex );
    M.measure();
    Box tmp = new Box(); tmp.setW( M.getW() );
    append( tmp );
  }


  /* ********************************************************************** */
  /* ********************************************************************** */
  /*         internal utility functions to implement TeX macros             */
  /* ********************************************************************** */
  /* ********************************************************************** */

 
  void greek( String s ) { // separate method in case we want symbol font here
    append( new CharBox( s, mathGreekFontIndex, ptSize, colorIndex ) );
    //append( new CharBox( s, fontIndex, ptSize, colorIndex ) );
  }

  void unicode( String s ) { 
    append( new CharBox( s, mathSymbolFontIndex, ptSize, colorIndex ) );
  }

  void mathfunction( String s ) {
    int fontIndex = this.fontIndex & 0xFC; // clears italic and bold bits
    append( new CharBox( s, fontIndex, ptSize, colorIndex ) );
  }
  
  void space( int points ) {
    Box spacer = new Box();
    spacer.setW( points );
    append( spacer );
  }



 
  void writeFigHeader( PrintStream ps ) {
    ps.println( "#FIG 3.2" );
    ps.println( "Portrait" );
    ps.println( "Center" );
    ps.println( "Metric" );
    ps.println( "A4" ); 
    ps.println( "100.0" );
    ps.println( "Single" );
    ps.println( "-2" );
    ps.println( "1200 2" );
  }
  
  
  static int count = 100; // number output files: hugo0.fig, hugo1.fig ...
  
  
  public void setFontIndex( int fontIndex ) {
    this.fontIndex = fontIndex;
  }
  
  public void setFontPtSize( int ptSize ) {
    this.ptSize = ptSize;
  }
  
  public void setColorIndex( int colorIndex ) {
    this.colorIndex = colorIndex;
  }
  
  public void parse( String s ) {
    tokenize( s );
  
    if (dump_tokens) {
      System.out.println( "tokens for input:\n" + s  );
      dumpTokens();
      System.out.println(); System.out.println();
    }
  
  }
  
  public void convertToFig( FigCompound compound ) {
    Box master = buildBox();
    master.convertToFig( compound );
  }
  
  public void Parse( String s ) {
    parse(s);
    Box master = buildBox();
    if (debug) master.dump( System.out );
  
    try {
      count ++;
      FileOutputStream fos = new FileOutputStream( "hugo" + count + ".fig" ); 
  
      FigCompound compound = new FigCompound();
      compound.setTrafo( new FigTrafo2D() );
      master.convertToFig( compound );
      compound.update_bbox();
      // compound.writeAsResource( new PrintWriter( fos ));
  
      FigWriter writer = new FigWriter();
      FigObjectList objectList = new FigObjectList();
      objectList.insert( compound );
      writer.write_fig31_file( new DataOutputStream(fos), true, objectList );
  
      fos.close(); 
    } catch( Throwable t ) { }
  
  
    //Box b1 = new CharBox( "Hugo", FigAttribs.FONT_TIMES_ROMAN, 20 );
    //Box b2 = new CharBox( "Bohnenstange", FigAttribs.FONT_COURIER, 12 );
    //System.out.println( b1 );
    //System.out.println( b2 );
  }
  
  
  
  public void setEnableDisplayBoxes( boolean b ) {
    display_boxes = b;
  }
  
  public void setEnableDumpTokens( boolean b ) {
    dump_tokens = b;
  }
  
  
  public static void msg( String s ) {
    System.out.println( s );
  }
  
  
  public static void usage() {
    msg( "java jfig.utils.SimpleLatexParser [options]" );
    msg( "where options include: " );
    msg( "-help               - this message " ); 
    msg( "-debug              - lots of diagnostic output" );
    msg( "-tokens             - dump tokens to stdout" );
    msg( "-boxes              - display the box outlines" );
    msg( "-italic <value>     - add <value> pts after italic chars" );
    System.exit(0);
  }
  
  
  public static void main( String argv[] ) {
    SimpleLatexParser slp = new SimpleLatexParser();
    for( int i=0; i < argv.length; i++ ) {
      if ("-tokens".equals( argv[i] )) slp.dump_tokens = true;
      if ("-boxes".equals( argv[i] ))  slp.display_boxes = true;
      if ("-debug".equals( argv[i] ))  slp.debug = true;
      if ("-italic".equals( argv[i] )) {
        slp.italic_correction = Integer.parseInt( argv[i+1] );
      }
      if ("-help".equals( argv[i] )) usage();
    }
  
  
    slp.setFontIndex( FigAttribs.FONT_TIMES_ROMAN );
    slp.setFontPtSize( 27 );
    slp.Parse( "Und es begab sich zu der Zeit, da Cyrhenius Landpfleger..." );
    slp.Parse( "Black, {\\red red}, {\\gold gold}, {\\green green}, {\\blue blue {\\black black inside} blue}!" );
    slp.Parse( "Short Commands, \\$ \\% \\\" aou\"a\"o\"uAOU\"A\"O\"Us\"s!" );
    slp.Parse( "Text und {\\tt Code mit {\\itt kursivem} Argument.}" );
    slp.Parse( "Das {\\bf Theorem} von {\\em Archimedes}." );
    slp.Parse( "{\\em\\red hendrich\\thinspace}{\\black \\atsign informatik.uni-hamburg.de}" );
    slp.Parse( "Normaler Text, {\\blue $w = \\sin ( \\alpha \\cdot \\pi /4 )$}" );
    slp.Parse( "{\\twentypt 20pt {\\fifteenpt 15pt }{\\twelvept 12pt }{\\tenpt 10pt }{\\eightpt 8pt }{\\fivept 5pt } and back to 20pt}" );
    slp.Parse( "Text mit Formel, {\\blue$w = \\sin( \\alpha\\cdot\\pi/4 )$}" );
  
    slp.setFontIndex( FigAttribs.FONT_HELVETICA );
    slp.setFontPtSize( 20 );
    slp.Parse( "Boole'sche Algebra: $a = b \\oplus c$ " );
    slp.Parse( "\\lessthan X\\greaterthan \\equals\\backslash\\$\\lbrace\\lbracket X\\rbracket\\rbrace X" );
     slp.Parse( "M\\emspace M $\\black\\clubsuit\\spadesuit {\\red\\heartsuit{}\\diamondsuit}$ \\telephone\\male\\ \\ \\female 42.59\\ \\euro, 15\\thinspace\\pound, 22.000\\thinspace\\pts" );
  
    slp.setFontIndex( FigAttribs.FONT_TIMES_ROMAN );
    slp.setFontPtSize( 30 );
    slp.Parse( "{\\darkgreen Pythagoras:} $a^{2} + b^{2} = c^{2}$" );
    slp.Parse( "$\\cos^{2}(x) + \\sin^{2}(x) = 1$" );
    slp.Parse( "{\\darkblue Euler formula:  $\\exp (i \\pi ) + 1 = 0$}" );
    slp.Parse( "{\\darkblue Euler formula:  $e^{\\thinspace i \\pi} + 1 = 0$}" );
    slp.Parse( "AbCdE: \\int $X_{j=3}^{2} \\cdot dx = 42.0$" );
    slp.Parse( "Formel1: $X_{i} = \\sum_{i=1}^{N_{0}} \\phi  \\xi  \\cdot 3$" );
    slp.Parse( "$\\prod_{i} \\sum_{j} \\Gamma_{\\Xi_{\\mu}} \\Rightarrow \\setZ \\bullet (TeX \\mp Word)$" );
    slp.Parse( "Hopfield-Dynamik: $S_{i} (t+1) = (1/N) \\ {\\fourtypt\\sum}_{j}\\  J_{ij}\\cdot S_{j} (t)$" );
    slp.Parse( "Hopfield-Lernen: $J_{ij} = (1/N) \\sum_{\\mu} \\xi_{i}^{\\mu} \\xi_{j}^{\\mu}$ " );
    slp.Parse( "${\\fourtypt\\int}^{5}_{0}\\ (x^{2}) {\\rm d}x = 1/3 x^{3}{\\fourtypt |}_{0}^{5} = (125/3) $" ); 
    slp.Parse( "$\\int^{3}_{x=0} a$" );
    slp.Parse( "IEEE 754: $value = (-1)^{s} \\cdot\\ 2^{exp-127} \\cdot\\ (1 + mantisse^{-23})$ ");
    slp.Parse( "Text (in) [Klammern] \\{und\\} mit\\_unterstrich" );
  
    System.exit( 0 );
  }
  
}
